/************************************************
*												*
*	EXTmacros.h									*
*												*	
*	DESCRIPTION:								*
*		All kinds of macros for ExTerra			*
*												*
*	Ofer Huberman, 2000.						*
*												*
************************************************/



#include <math.h>
#include <stdlib.h>

#define EXT_MACROS_H
double EXT_VAR_NORMALIZE;
/****************************************************
*													*
*	Description:									*
*		Mathematical routines and symbols.			*
*													*
*	EXT_SQR(p):										*
*		Returns p squared.							*
*													*
*****************************************************/

#define EXT_SQR(p)  ( (p) * (p) )
#define EXT_CUBIC(p)  ( (p) * (p) * (p) )
#define EXT_SQRT(p) ( sqrt( (p) ) )

#define EXT_SPHEREx(theta, fee, radius)	( cos( (theta) ) * sin( (fee) ) * (radius) )
#define EXT_SPHEREy(theta, fee, radius)	( sin( (theta) ) * sin( (fee) ) * (radius) )
#define EXT_SPHEREz(fee, radius)		( cos( (fee) ) * (radius) )


#define EXT_PI	3.141592653
#define EXT_RAD	6.283185307
#define EXT_90	1.570796326


/****************************************************
*													*
*	Description:									*
*		Point handling macros.						*
*													*
*	EXT_SET_TO_ORIGIN(p):							*
*		Sets point (p) to origin.					*
*													*
*	EXT_POINT_COPY3(p, vx, vy, vz)					*
*		Set point (p) to given x, y, and z.			*
*													*
*	EXT_POINT_TRANSLATE3(p, vx, vy, vz)				*
*		Translates point (p) by vx, vy, and vz.		*
*													*
*	EXT_POINT_AS_PARAM(p)							*
*		Returns point in a x,y,z format.			*
*													*
*****************************************************/

#define EXT_DISTANCE2_NS(a,b) ( EXT_SQR((a).x - (b).x) + EXT_SQR((a).y - (b).y) + EXT_SQR((a).z - (b).z) )

#define EXT_DISTANCE_ORIGIN_NS(a) ( EXT_SQR((a).x) + EXT_SQR((a).y) + EXT_SQR((a).z) )

#define EXT_POINT_DISTANCE(a,b) ( sqrt( EXT_DISTANCE2_NS( (a), (b) ) ) )
#define EXT_DISTANCE_ORIGIN(a) ( sqrt(EXT_DISTANCE_ORIGIN_NS( (a) ) ) )

#define EXT_SET_TO_ORIGIN(p) {	(p).x = 0.0; \
								(p).y = 0.0; \
								(p).z = 0.0; }
#define EXT_SET_TO_ORIGINp(p) {	(p)->x = 0.0; \
								(p)->y = 0.0; \
								(p)->z = 0.0; }


#define EXT_POINT_COPY3(p,vx,vy,vz) {	(p).x = vx; \
										(p).y = vy; \
										(p).z = vz; }

#define EXT_POINT_COPY(p,s) {	p.x = s.x; \
								p.y = s.y; \
								p.z = s.z; }


#define EXT_POINT_TRANSLATE3(p, vx, vy, vz) {	(p).x += vx; \
											(p).y += vy; \
											(p).z += vz; }

#define EXT_POINT_TRANSLATE(p, v) { (p).x += (v).x; \
									(p).y += (v).y; \
									(p).z += (v).z; }

#define EXT_POINT_AS_PARAM(p) (p).x, (p).y, (p).z
#define EXT_PARAM(p) (p).x, (p).y, (p).z

#define EXT_ORIGIN 0.0, 0.0, 0.0

/****************************************************
*													*
*	Description:									*
*		Vector handling routines.					*
*													*
*	EXT_VECTOR_MAGNITUDE(p):						*
*		Returns magnitude of vector p.				*
*													*
*	EXT_VECTOR_MAGNITUDE_NS(p):						*
*		Returns squared magnitude of vector p		*
*													*
*****************************************************/

#define EXT_VECTOR_MAGNITUDE_NS(p) ( EXT_SQR( (p).x ) + EXT_SQR( (p).y ) + EXT_SQR( (p).z ) )

#define EXT_VECTOR_MAGNITUDE(p) ( sqrt( EXT_VECTOR_MAGNITUDE_NS( (p) ) ) )

#define EXT_VECTOR_TO_ORIGIN(p, v) {	(v).x = -(p).x;\
										(v).y = -(p).y;\
										(v).z = -(p).z; }

#define EXT_VECTOR_MULTIPLY2(t, v, s) {	(t).x = (v).x * (s); \
									(t).y = (v).y * (s); \
									(t).z = (v).z * (s); }

#define EXT_VECTOR_MULTIPLY(v, s) {	(v).x *= (s); \
									(v).y *= (s); \
									(v).z *= (s); }

#define EXT_VECTOR_DIVIDE(v, s) {	(v).x /= s; \
									(v).y /= s; \
									(v).z /= s; }


#define EXT_VECTOR_CROSS_PRODUCT(v1, v2, v3) {	(v3).x = (v1).y * (v2).z - (v1).z * (v2).y; \
												(v3).y = (v1).x * (v2).z - (v1).z * (v2).x; \
												(v3).z = (v1).x * (v2).y - (v1).y * (v2).x; }

#define EXT_VECTOR_DOT_PRODUCT(v1, v2) ( (v1).x * (v2).x + (v1).y * (v2).y + (v1).z * (v2).z )

#define EXT_VECTOR_NORMALIZE(v) {	EXT_VAR_NORMALIZE = EXT_VECTOR_MAGNITUDE( (v) ); \
									EXT_VECTOR_DIVIDE( (v), EXT_VAR_NORMALIZE); }

#define EXT_VECTOR_BETWEEN_POINTS(p1, p2, v) {	(v).x = (p1).x - (p2).x; \
												(v).y = (p1).y - (p2).y; \
												(v).z = (p1).z - (p2).z; \
												EXT_VECTOR_NORMALIZE( (v) ) ; }

#define EXT_VECTOR_AVERAGE(v1, v2, v3) {	(v3).x = ((v1).x + (v2).x ) / 2; \
											(v3).y = ((v1).y + (v2).y ) / 2; \
											(v3).z = ((v1).z + (v2).z ) / 2; }

/****************************************************
*													*
*	Description:									*
*		Random and noise functions.					*
*													*
*	EXT_RND_BELL(p):								*
*		Returns random number between p and -p.		*
*													*
*****************************************************/

#define EXT_RND_BELL(p) ( (rand() * (p) / RAND_MAX) - (rand() * (p) / RAND_MAX ) )
#define KM
