#include <stdio.h>			/* Standard I/O header	*/
#include <gl/glut.h>		
#include <gl/glu.h>
#include <gl/gl.h>			/* The Holy Triplet		*/

#define WIN32_LEAN_AND_MEAN
	#include <windows.h>

#define HF_DIM 2049 /* Dimensions of height-field */

int current;

#define EXT_OGL

#include "EXTbasetypes.h"
#include "EXToglmacros.h"
#include "EXTmacros.h"	/* Personal macros and other aiding functions	*/

#include "heightfield.h"
#include "quadtree.h"
#include "task.h"

#define ABS(x)		( (x) > 0 ? (x) : (-(x) ) )

float C;
float clouds;
float zangle;
float old_c;
long old_time;
long DESIRED_TRIS =	3600;
int Shots;
int sky;
int alods = 1;
char text[120];

HEIGHTFIELD	*data[2];
QUADTREE	*skeleton[2];

long tl [5];
double cl[5];

EXTcamera camera;			/* Our camera	*/

unsigned char *texture;
unsigned int texture_list[10];

void mainTakeShot ()
{
	long i, end;
	unsigned char *dat;
	char filename[40];
	FILE *file;

	sprintf (filename, "SHOT%d.raw", Shots);
	
	end = 1024 * 768 * 3;
	
	dat = calloc (end, sizeof (unsigned char) );
	glReadPixels (0, 0, 1024, 768, GL_RGB, GL_UNSIGNED_BYTE, dat);

	file = fopen (filename, "wb");

	for (i = 0; i < end; i++)
		fputc (dat[i], file);

	fclose (file);
	free (dat);
	free (filename);

	Shots++;
}

void mainDrawSky ()
{
	float i;

	i = sqrt (EXT_SQR(camera.clip_far) / 2.0) - 200;

	glTranslated (camera.location.x, camera.location.y, camera.location.z);	

	glDisable (GL_DEPTH_TEST);
	glDisable (GL_FOG);

	glMatrixMode (GL_TEXTURE);		/* Setup texture matrix */
		glLoadIdentity();

	glRotatef(clouds, 0.0, 0.433, 0.9);

	glCullFace (GL_BACK);
	
	glBegin (GL_TRIANGLE_FAN);
	{
		glTexCoord2f (0.5, 0.5);
		glColor3f (0.1, 0.1, 0.7);
		glVertex3d (0, 0, 600.0);

		glTexCoord2f (0.0, 0.0);
		glColor3f (0.85, 0.85, 1.0);
		glVertex3f (-i, -i, 0);
		glTexCoord2f (0.0, 1.0);
		glColor3f (0.85, 0.85, 1.0);
		glVertex3f (-i, i, 0);
		glTexCoord2f (1.0, 1.0);
		glColor3f (0.85, 0.85, 1.0);
		glVertex3f (i, i, 0);
		glTexCoord2f (1.0, 0.0);
		glColor3f (0.85, 0.85, 1.0);
		glVertex3f (i, -i, 0);
		glTexCoord2f (0.0, 0.0);
		glColor3f (0.85, 0.85, 1.0);
		glVertex3f (-i, -i, 0);
	}
	glEnd ();

	glCullFace (GL_FRONT);
	
	glMatrixMode (GL_TEXTURE);		/* Setup texture matrix */
		glLoadIdentity();
		glRotatef(clouds, 0.0, -0.433, 0.9);

	glBegin (GL_TRIANGLE_FAN);
	{
		glTexCoord2f (0.5, 0.5);
		glColor3f (1.0, 0.5, 0.2);
		glVertex3d (0, 0, -600.0);

		glTexCoord2f (0.0, 0.0);
		glColor3f (0.85, 0.85, 1.0);
		glVertex3f (-i, -i, 0);
		glTexCoord2f (0.0, 1.0);
		glColor3f (0.85, 0.85, 1.0);
		glVertex3f (-i, i, 0);
		glTexCoord2f (1.0, 1.0);
		glColor3f (0.85, 0.85, 1.0);
		glVertex3f (i, i, 0);
		glTexCoord2f (1.0, 0.0);
		glColor3f (0.85, 0.85, 1.0);
		glVertex3f (i, -i, 0);
		glTexCoord2f (0.0, 0.0);
		glColor3f (0.85, 0.85, 1.0);
		glVertex3f (-i, -i, 0);
	}
	glEnd ();

	glLoadIdentity ();

	glScaled(14.0 / HF_DIM , 14.0 / HF_DIM, 0.1);

	glMatrixMode (GL_MODELVIEW);
		glTranslated (-camera.location.x, -camera.location.y, -camera.location.z);

	glEnable (GL_FOG);
	glEnable (GL_DEPTH_TEST);

	clouds += 0.04;
}

void mainRenderScene(void)	/* GLUT rendering function */
{
	int i;
	EXTpoint3 look_at;
	long time;
	long nTris;
	
	time = GetTickCount ();

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);	/* Clear viewport */

	camera.fwd.x = cos(camera.angle + zangle) * cos(camera.pitch);
	camera.fwd.y = sin(camera.angle + zangle) * cos(camera.pitch);
	camera.fwd.z = sin(camera.pitch);

	look_at.x = camera.location.x + camera.fwd.x;
	look_at.y = camera.location.y + camera.fwd.y;
	look_at.z = camera.location.z + camera.fwd.z;

	glMatrixMode(GL_MODELVIEW);							/* Set model-view matrix */
		glLoadIdentity();
		glRotated (camera.roll, 0.0, 0.0, 1.0);
		
		gluLookAt(camera.location.x, camera.location.y, camera.location.z, EXT_POINT_AS_PARAM(look_at), 0.0, 0.0, 1.0);	

	if (sky)
		mainDrawSky ();
	
	glScaled (1.0, 1.0, 0.7);

	quadRenderTree (skeleton[0], data[0], &camera, C);	
	
	if (alods)
	{
		nTris = 0;

		if (nTris != DESIRED_TRIS)
		{
			C += (DESIRED_TRIS - skeleton[0]->nTris) / 5000.0;
			if (C <= 0)
				C = 1e-5;
		}
		C = ( (C * 3.0) + (cl[0] * 2.0) + (cl[1] * 1.0) ) / 6.0;

		old_c = C;		
		
		tl [4] = tl[3];
		tl [3] = tl[2];
		tl [2] = tl[1];
		tl [1] = tl[0];
		tl [0] = DESIRED_TRIS - nTris + 1;

		cl [4] = cl[3];
		cl [3] = cl[2];
		cl [2] = cl[1];
		cl [1] = cl[0];
		cl [0] = C;

		if (ABS(C - old_c) > 0.3)
			C = (C + old_c + old_c) / 3.0;

		old_c = C;
	}
	
	time = GetTickCount () - time;
	
	for (i = 0; i < 120; i++)
		text[i] = 0;

	if (time != 0)
		sprintf (text, "# Fans: %d. Delta: %d. FP/S: %5.1f.", skeleton[0]->nTris, skeleton[0]->nTris - DESIRED_TRIS, 2000.0  / (time + old_time));

	glutSetWindowTitle (text);
	text[119] = 0;	
	
	glLoadIdentity ();
		gluLookAt (0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 1.0, 0.0);
		glTranslatef (35.0f, -30.0f, 40.0f);

	glRasterPos2f (0.0f, 0.0f);	
	glDisable (GL_TEXTURE_2D);
	glDisable (GL_DEPTH_TEST);
	glColor3ub (0, 0, 0);
	
	i = 0;
	while (text[i])
	{
		glutBitmapCharacter (GLUT_BITMAP_HELVETICA_10, text[i]);
		i++;
	}

	glEnable (GL_TEXTURE_2D);
	glEnable (GL_DEPTH_TEST);	

	old_time = time;
	glutSwapBuffers ();
}

void mainRenderTringulation ()
{
	glClearColor (1.0, 1.0, 1.0, 1.0);

	EXT_OGL_LINE;

	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glEnable (GL_LIGHTING);

	glMatrixMode (GL_PROJECTION);
		glLoadIdentity ();
		glOrtho (0.0, skeleton [0]->dim, 0.0, skeleton [0]->dim, 1.0, 7000.0);

	glMatrixMode (GL_MODELVIEW);
		glLoadIdentity ();	
		gluLookAt (0.0, 0.0, 500.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);

	quadRenderTree (skeleton [0], data [0], &camera, C);

	glutSwapBuffers ();
}

void mainAnimateFunc (void)
{
	if (ABS(camera.speed) > 0)
	{
		camera.location.x += 0.5 * camera.speed * (cos(camera.fwd_a) * cos(camera.pitch) );
		camera.location.y += 0.5 * camera.speed * (sin(camera.fwd_a) * cos(camera.pitch) );
		camera.location.z += 0.5 * camera.speed * sin(camera.pitch);

		camera.angle -= sin (camera.roll / 180.0 * EXT_PI) * 0.016;
		camera.fwd_a -= sin (camera.roll / 180.0 * EXT_PI) * 0.016;
	}

	if (camera.pitch > EXT_PI)
	{
		camera.angle += EXT_PI;
		camera.fwd_a += EXT_PI;
	}
	if (camera.pitch < -EXT_PI)
	{
		camera.angle += EXT_PI;
		camera.fwd_a += EXT_PI;
	}

	glutPostRedisplay();

	SetCursorPos (400, 400);
	return;
}

void mainCallbackFunc(unsigned char key, int x, int y) /* GLUT keyboard-callback function */
{
	switch (key)
	{
		case 'D':
			zangle -= 0.05;
			break;
		case 'A':
			zangle += 0.05;
			break;
		case 'S':
			zangle = 0;
			break;
		case 'w':
			camera.speed += 0.1;
 			break;
		case 's':
			camera.speed -= 0.1;
			break;
		case 'a':
			camera.angle += 0.015 * cos(EXT_PI * camera.roll / 180.0);
			camera.fwd_a += 0.015 * cos(EXT_PI * camera.roll / 180.0);
			camera.pitch += 0.015 * sin(EXT_PI * camera.roll / 180.0);
			break;
		case 'd':
			camera.angle -= 0.015 * cos(EXT_PI * camera.roll / 180.0);
			camera.fwd_a -= 0.015 * cos(EXT_PI * camera.roll / 180.0);
			camera.pitch -= 0.015 * sin(EXT_PI * camera.roll / 180.0);
			break;
		case 'q':
			camera.roll -= 2.0;
			break;
		case 'e':
			camera.roll += 2.0;
			break;
		case 'c':
			camera.angle -= 0.015 * sin(EXT_PI * camera.roll / 180.0);
			camera.fwd_a -= 0.015 * sin(EXT_PI * camera.roll / 180.0);
			camera.pitch += 0.015 * cos(EXT_PI * camera.roll / 180.0);
			break;
		case 'v':
			camera.angle += 0.015 * sin(EXT_PI * camera.roll / 180.0);
			camera.fwd_a += 0.015 * sin(EXT_PI * camera.roll / 180.0);
			camera.pitch -= 0.015 * cos(EXT_PI * camera.roll / 180.0);
			break;
		case 27:
		case 'p':
			taskNew ("Cleaning up");
				quadFreeMemory (skeleton[0]);
				hfFreeMemory (data[0]);
				glutLeaveGameMode ();
			taskOk ();
			ShowCursor (1);
			exit(1);
		case '+':
			if (alods)
				DESIRED_TRIS += 50;
			else
				C *= 1.1;
			break;
		case '-':
			if (alods)
				DESIRED_TRIS -= 50;
			else
				C /= 1.1;
			break;		
		case '1':
			skeleton[0]->mul_roughness = 0;
			EXT_OGL_FILL;
			glEnable (GL_TEXTURE_2D);
			break;
		case '2':
			skeleton[0]->mul_roughness = 1;
			EXT_OGL_LINE;			
			break;
		case '3':
			sky = 1;
			break;
		case '4':
			sky = 0;
			break;
		case '\\':
			mainTakeShot ();
			break;
		case '0':
			alods = 1;
			break;
		case '9':
			alods = 0;
			break;
	}
	
	x += y; /* Avoid unreferenced formal parameter warning */
}

void mainPassiveFunc (int x, int y)
{
	camera.roll += ( (double) x - (double) 400) / 23.0;
	
	camera.angle -= ( ( (double) y - (double) 400) / 2000.0) * sin(EXT_PI * camera.roll / 180.0);
	camera.fwd_a -= ( ( (double) y - (double) 400) / 2000.0) * sin(EXT_PI * camera.roll / 180.0);
	camera.pitch += ( ( (double) y - (double) 400) / 2000.0) * cos(EXT_PI * camera.roll / 180.0);
}

void mainMotionFunc (int x, int y)
{
	if (GetAsyncKeyState (VK_LCONTROL) )
		camera.roll += ( (double) x - (double) 400) / 23.0;
	if (GetAsyncKeyState (VK_LSHIFT) )
		camera.angle -= ( (double) x - (double) 400) / 2000.0;
	else
		camera.speed -= ( (double) y - (double) 400) / 450;
		
}

void mainMouseFunc (int b, int s, int x, int y)
{
	x = s + y + b;
}

void mainGetTexture ()
{
	long i, end = sizeof (char) * 3 * 256 * 256;
	FILE *stream = fopen ("texture.raw", "rb");
	
	texture = (unsigned char *) malloc (end);

	glGenTextures (10, texture_list);

	for (i = 0; i < end; i++)
		texture[i] = fgetc (stream);

	glBindTexture	(GL_TEXTURE_2D, 1);
	
	glPixelStorei (GL_UNPACK_ALIGNMENT, 1); 
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE); 

	glTexImage2D (GL_TEXTURE_2D, 0, 1, 256, 256, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, texture);

	free (texture);
}

void mainReshapeFunc (int x, int y)
{
	glViewport (0, 0, x, y);
}

int main(int arg_c, char *arg_v[])
{
	int resolution;
	float	fog_color [] = { 0.5, 0.5, 0.5, 0.5 };
	
	tl[0] = tl[1] = DESIRED_TRIS + 1;
	tl[2] = tl[3] = DESIRED_TRIS + 1;
	tl[4] = DESIRED_TRIS + 1;
	cl[0] = cl[1] = cl[2] = cl[3] = cl[4] = 2.0;

	zangle = 0;
	clouds = 0;	
	Shots = 0;
	alods = 1;

	printf ("- XTerra: Graphics Engine for Real-Time Visualization of Height-Fields -\n");
	printf ("\nImplementation of 'Generation of Multiple Levels of Detail for Height-Fields', by S. Rottger et al.\n");
	printf ("Developed by Ofer Huberman, 2000.\n");
	old_c = C = 0.2;
	sky = 1;
	old_time = 0;
	
	printf ("\n1) 640x480\n2) 800x600\n3) 1024x768\n4) Windowed\n");
	printf ("Choose desired screen resolution: ");
	scanf ("%d", &resolution);

	taskNew ("Initializing data-types");
		skeleton[0] = quadInit (HF_DIM);		
	taskOk ();

	taskNew ("Allocating memory for height-field data");
		data[0]	= hfAllocateMemory (HF_DIM, HF_DIM);
	taskOk ();	
	
	taskNew ("Resetting height-field");
		hfResetField (data[0]);			
	taskOk ();

	taskNew ("Retrieving height-field");
 		hfRetrieveField (data[0], "x1.raw");
	taskOk ();

	//hfCalculateHistogramUB (data [0]);
	//scanf ("%d");

	taskNew ("Processing height-field");
		quadProcessTree (skeleton[0], data[0]);
	taskOk ();

	taskNew ("Light-shading map");
		hfLightUpField (data[0]);
	taskOk ();	

	taskNew ("Setting up GLUT");
	
		glutInit(&arg_c, arg_v);
		glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
		glutInitWindowPosition(0, 0);
		glutInitWindowSize(800, 600);
		
		if (resolution == 4)
		{
			glutCreateWindow ("XTerra");
			glViewport (0, 0, 800, 600);
		}
		else
		{
			if (resolution == 3)
			{	
				glutGameModeString ("1024x768:16@60");
				glViewport (0, 0, 1024, 768);
			}
			if (resolution == 2)
			{
				glutGameModeString ("800x600:16@60");
				glViewport (0, 0, 800, 600);
			}
			if (resolution == 1)
			{
				glutGameModeString ("640x480:16@60");
				glViewport (0, 0, 640, 480);
			}
			glutEnterGameMode ();
		}

		glutDisplayFunc		(mainRenderScene);
	//	glutDisplayFunc		(mainRenderTringulation);	
		glutKeyboardFunc	(mainCallbackFunc);	/* Setup GLUT callback functions */
		glutIdleFunc		(mainAnimateFunc);
		glutPassiveMotionFunc (mainPassiveFunc);
		glutMotionFunc		(mainMotionFunc);

	taskOk ();

	taskNew ("Setting up OpenGL");
	
		glEnable (GL_TEXTURE_2D);
		glEnable (GL_DEPTH_TEST);
		glEnable (GL_CULL_FACE);	/* Set OpenGL attribs */
		glCullFace (GL_FRONT);
		
		glClearColor(0.05, 0.05, 0.4, 0.6);
		
		glLineWidth(1.0);
		
		taskNew ("Initializing camera");
			EXT_POINT_COPY3 (camera.location, 400.0, 400.0, 270.0);
			camera.roll			= 0.0f;		/* Setup camera properties */
			camera.angle		= 0.0f;
			camera.fwd_a		= 0.0f;
			camera.pitch		= 0.0f;
			camera.speed		= 0.1f;
			camera.clip_near	= 0.2f;
			camera.clip_far		= 2500.0f;
			camera.fov			= 90.0f;
			camera.half_fov		= 0.2 + (camera.fov / (2.0 * (180.0 / EXT_PI) ) );	
		taskOk ();

		glEnable (GL_FOG);
		glFogfv (GL_FOG_COLOR, fog_color);
		glFogf (GL_FOG_START, 0.0);
		glFogf (GL_FOG_END, 2750.0);
		glFogf (GL_FOG_DENSITY, 0.0016);

		glMatrixMode(GL_PROJECTION);	/* Setup projection matrix */
			glLoadIdentity();
			gluPerspective(camera.fov, 1.3333, camera.clip_near, camera.clip_far);			

		glPushMatrix();

		taskNew ("Retrieving texture");
			mainGetTexture ();	
		taskOk ();		
		
		glMatrixMode(GL_TEXTURE);		/* Setup texture matrix */
			glLoadIdentity();
			glScaled(14.0 / HF_DIM , 14.0 / HF_DIM, 0.1);

	taskOk ();

	printf("\nGL Driver Vendor: %s\n", glGetString (GL_VENDOR) );
	ShowCursor (0);

	glutMainLoop(); /* Go GLUT! */
	return 0;
}