/***********************z************************************************************
**																					*
**	Header file for The Division Bell:												*
**																					*
**	Generator of Multiple Levels of Detail for Fast Visualization of Height Fields	*
**																					*
**	Written by Ofer Huberman, April, May, June, 2000.								*
**																					*
**	Based on 'Real-Time Generation of Continuous Level of Detail for Height-Fields	*
**	By S. Rottger, et al.															*
**																					*
************************************************************************************/

#include "heightfield.h"

/****************************************************
*													*
*	QUADNODE STRUCTURE:								*
*													*
*	DESCRIPTION: Describes one node in the tree.	*
*													*
*	VARIABLES:										*
*													*
*	roughness	- Roughness of current node			*
*	status		- status of current node			*
*	lightness	- lighting factor					*
*													*
****************************************************/

typedef struct _quadnode
{
	unsigned char roughness;
	unsigned char status;
} QUADNODE;

/****************************************************
*													*
*	QUADTREE STRUCTURE:								*
*													*
*	DESCRIPTION: Describes one quad-tree.			*
*													*
*	VARIABLES:										*
*													*
*	**nodeMatrix	- Two dimensional array of nodes*
*	dim				- Dimensions of tree			*
*	center_and_...	- (dim + 1) : location of root	*
*					  ---------						*
*						  2							*
*													*
****************************************************/

typedef struct _quadtree
{
	QUADNODE		**nodeMatrix;
	INDEXType		dim;
	INDEXType		center_and_width;
	float			mul_roughness;
	long			nTris;
} QUADTREE;

void	*quadInit				(int pDimensions);
void	quadFreeMemory			(QUADTREE *pTree);
void	quadProcessTree			(QUADTREE *pTree, HEIGHTFIELD *pField);
void	quadRenderTree			(QUADTREE *pTree, HEIGHTFIELD *pField, EXTcamera *pCamera, float pC);